package healthcheck

import (
	"context"
	"fmt"
	"net"
	"net/http"
	"net/http/httptest"
	"strconv"
	"strings"
	"testing"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/testutils"
)

func makeFakeHTTPService(route, data string, statusCode int) (*httptest.Server, string, int) {
	server := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if r.URL.Path != route {
			return
		}

		w.WriteHeader(statusCode)
		fmt.Fprint(w, data)
	}))

	parts := strings.Split(server.URL, ":")
	ip := parts[1][2:]
	port, err := strconv.ParseInt(parts[2], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return server, ip, int(port)
}

func runFakeTCPService(l net.Listener, data string) {
	for {
		conn, err := l.Accept()
		if err != nil {
			// https://grokbase.com/t/gg/golang-nuts/131k81bse0/go-nuts-use-of-closed-network-connection-error-from-tcpconn-read
			netOpError, ok := err.(*net.OpError)
			if ok && netOpError.Err.Error() == "use of closed network connection" {
				return
			}

			log.Warn(err)
			return
		}
		go func(conn net.Conn) {
			defer conn.Close()
			var res []byte
			_, err := conn.Read(res)
			if err != nil {
				log.Fatalln(err)
			}
			_, err = conn.Write([]byte(data))
			if err != nil {
				log.Fatalln(err)
			}
		}(conn)
	}
}

func makeFakeTCPService(data string) (net.Listener, int) {
	listener, err := net.Listen("tcp", "127.0.0.1:0")
	if err != nil {
		log.Fatalln(err)
	}
	go runFakeTCPService(listener, data)

	addr := listener.Addr().String()

	port, err := strconv.ParseInt(strings.Split(addr, ":")[1], 10, 64)
	if err != nil {
		log.Fatalln(err)
	}

	return listener, int(port)
}

func makeFakeServices(healthchecks map[string]healthcheckI) (*common.Inventory, map[string]float64, []*httptest.Server, []net.Listener) {
	expected := map[string]float64{}
	httpServers := make([]*httptest.Server, 0, 5)
	tcpListeners := make([]net.Listener, 0, 5)

	services := make(common.ServicesMap)
	inventory := &common.Inventory{
		Services: services,
	}

	httpServer, ip, port := makeFakeHTTPService(healthchecks["account"].(*HTTPCheck).Route, "", 200)
	services["account"] = []*common.Service{
		{
			ID:   "OPENIO-account-0",
			IP:   ip,
			Port: port,
		},
	}
	expected["OPENIO-account-0"] = float64(1)
	httpServers = append(httpServers, httpServer)

	tcpListener, port := makeFakeTCPService("OK \n")
	services["meta0"] = []*common.Service{
		{
			ID:   "OPENIO-meta0-0",
			IP:   "127.0.0.1",
			Port: port,
		},
	}
	expected["OPENIO-meta0-0"] = float64(1)
	tcpListeners = append(tcpListeners, tcpListener)

	tcpListener, port = makeFakeTCPService("OK \n")
	services["beanstalkd"] = []*common.Service{
		{
			ID:   "OPENIO-beanstalkd-0",
			IP:   "127.0.0.1",
			Port: port,
		},
	}
	expected["OPENIO-beanstalkd-0"] = float64(1)
	tcpListeners = append(tcpListeners, tcpListener)

	tcpListener, port = makeFakeTCPService("KO \n")
	services["beanstalkd"] = append(services["beanstalkd"], &common.Service{
		ID:   "OPENIO-beanstalkd-1",
		IP:   "127.0.0.1",
		Port: port,
	})
	expected["OPENIO-beanstalkd-1"] = float64(0)
	tcpListeners = append(tcpListeners, tcpListener)

	httpServer, ip, port = makeFakeHTTPService(healthchecks["keystone"].(*HTTPCheck).Route, "", 401)
	services["keystone"] = []*common.Service{
		{
			ID:   "OPENIO-keystone-0",
			IP:   ip,
			Port: port,
		},
	}
	expected["OPENIO-keystone-0"] = float64(1)
	httpServers = append(httpServers, httpServer)

	return inventory, expected, httpServers, tcpListeners
}

func TestHealtcheckModule(t *testing.T) {
	t.Parallel()
	_log.LogLevel = _log.INFO
	_ = _log.Setup()

	factory := NewFactory("", "../healthchecks.yml").(*healthcheckDaemonFactory)
	if err := factory.Reload(); err != nil {
		t.Fatalf("Reload error: %s", err)
	}

	inventory, expected, httpServers, tcpListeners := makeFakeServices(factory.healthchecks)
	for _, httpServer := range httpServers {
		defer httpServer.Close()
	}
	for _, tcpListener := range tcpListeners {
		defer tcpListener.Close()
	}

	m := common.MakeMetricData()
	ctx := context.TODO()

	daemonParams := factory.Daemons(inventory)
	for _, params := range daemonParams {
		d := factory.New(params)

		if err := d.Init(ctx); err != nil {
			t.Fatalf("Init error: %s", err)
		}

		if err := d.Collect(ctx, m); err != nil {
			t.Fatalf("Collect error: %s", err)
		}

		if err := d.Close(); err != nil {
			t.Fatalf("Close error: %s", err)
		}
	}

	mdt := testutils.NewMetricDataTest()
	for service, expectedValue := range expected {
		mdt.ForMetricWithLabels(
			testutils.NewMetricAssertion().Equals(expectedValue),
			"probe_success",
			"service", service)
	}
	mdt.Check(t, m)
}
